// Tide Protocol - Infrastructure for the Personal Data economy
// Copyright (C) 2019 Tide Foundation Ltd
// 
// This program is free software and is subject to the terms of 
// the Tide Community Open Source License as published by the 
// Tide Foundation Limited. You may modify it and redistribute 
// it in accordance with and subject to the terms of that License.
// This program is distributed WITHOUT WARRANTY of any kind, 
// including without any implied warranty of MERCHANTABILITY or 
// FITNESS FOR A PARTICULAR PURPOSE.
// See the Tide Community Open Source License for more details.
// You should have received a copy of the Tide Community Open 
// Source License along with this program.
// If not, see https://tide.org/licenses_tcosl-1-0-en

import BigInt from 'big-integer';
import Hash from './hash';
import { getBig, getArray, getLittle } from './bnInput';
import Utils from './utils';
import C25519Point from './c25519Point';
import C25519Key from './c25519Key';

/**
 * @param {Uint8Array|string} message 
 * @param {C25519Key} key
 * @param {string} sigScheme
 * @returns {Uint8Array}
 */
export function sign(message, key, sigScheme) {
    var buff = typeof message === 'string' ? Buffer.from(message, 'utf-8') : message;
    if (sigScheme == "ecDSA"){
        var [r, s] = signEc(getM(buff), key);
        return map(r, s);
    }
    else if (sigScheme == "edDSA"){
        return signEd(buff, key); // fyi: mapEd is called and returned from inside signEd
    }
}

/**
 * @param {Uint8Array|string} message
 * @param {Uint8Array} signature
 * @param {C25519Key} key
 * @param {string} sigScheme
 */
export function verify(message, signature, key, sigScheme) {
    if (!signature || signature.length != 96)
        return false;

    var buff = typeof message === 'string' ? Buffer.from(message, 'utf-8') : message;
    if (sigScheme == "ecDSA"){
        var [r, s] = unmap(signature);
        return verifyEc(getM(buff), r, s, key);
    }
    else if (sigScheme == "edDSA"){
        var [rx, ry, s] = unmapEd(signature);
        return verifyEd(rx, ry, s, buff, key);
    }
    
}

/**
 * @param {BigInt.BigInteger} r
 * @param {BigInt.BigInteger} s
 */
export function map(r, s) {
    var bufferR = getArray(r);
    var bufferS = getArray(s);

    var buffer = Buffer.alloc(64);
    buffer.set(bufferR, 32 - bufferR.length);
    buffer.set(bufferS, 64 - bufferS.length);

    return buffer;
}
/**
 * @param {BigInt.BigInteger} r_x
 * @param {BigInt.BigInteger} r_y
 * @param {BigInt.BigInteger} s
 */
 export function mapEd(r_x, r_y, s) {
    var r_xBuff = r_x.toArray(256).value;
    while(r_xBuff.length < 32){r_xBuff.unshift(0);} // pad if array < 32
    r_xBuff.reverse(); // to LE
    
    var r_yBuff = r_y.toArray(256).value;
    while(r_yBuff.length < 32){r_yBuff.unshift(0);} // pad if array < 32
    r_yBuff.reverse(); // to LE

    var sBuff = s.toArray(256).value;
    while(sBuff.length < 32){sBuff.unshift(0);} // pad if array < 32
    sBuff.reverse(); // to LE

    var buffer = Buffer.alloc(96);
    buffer.set(r_xBuff, 32 - r_xBuff.length);
    buffer.set(r_yBuff, 64 - r_yBuff.length);
    buffer.set(sBuff, 96 - sBuff.length);

    return buffer;
}
/** @param {Uint8Array} signature */
export function unmap(signature) {
    return [getBig(signature.slice(0, 32)), getBig(signature.slice(32))];
}

/**
 * @param {Uint8Array} m
 * @param {C25519Key} key
 */
function signEd(m, key)
{
    var r = Utils.random(BigInt.one, C25519Point.n.subtract(BigInt.one));
    var R = C25519Point.g.times(r);
    var R_b = toRawBytes(R);

    var pub = toRawBytes(key.y);  // note that y constructor actually refers to public point of key.x. change this in future.
    var h = getLE(Hash.sha512Buffer(concatBytes(R_b, pub, m))).mod(C25519Point.n); // LE (hash (R || P || msg)) % n
    var s = (r.add(h.multiply(key.x))).mod(C25519Point.n); // (r + h * priv) % n

    return mapEd(R.x, R.y, s);
}

/**
 * @param {Uint8Array} signature
 */
function unmapEd(signature){
    return [getLittle(signature.slice(0, 32)), getLittle(signature.slice(32, 64)), getLittle(signature.slice(64))];
}

/**
 * @param {BigInt.BigInteger} rx
 * @param {BigInt.BigInteger} ry
 * @param {BigInt.BigInteger} s
 * @param {Uint8Array} message
 * @param {C25519Key} key
 */
function verifyEd(rx, ry, s, message, key){
    var R = new C25519Point(rx, ry);
    var h = getLE(Hash.sha512Buffer(concatBytes(toRawBytes(R), toRawBytes(key.y), message))).mod(C25519Point.n); // LE (hash (R || P || msg)) % n

    var point1 = C25519Point.g.times(s);
    var point2 = R.add(key.y.times(h))
    
    return point1.x.equals(point2.x) && point1.y.equals(point2.y);
}
/**
 * @param {Uint8Array} bytes 
 */
function getLE(bytes){
    return BigInt.fromArray(Array.from(bytes).reverse(), 256, false);
}

/**
 * 
 * @param {C25519Point} point 
 * @returns {Uint8Array}
 */
function toRawBytes(point) {
    const bytes = numberTo32BytesLE(point.y);
    if(point.x.and(BigInt.one).equals(BigInt.one)){
        var mask = 128;
        bytes[31] |= mask;
    }else{
        var mask = 127;
        bytes[31] &= mask;
    }
    return bytes;
}

/**
 * 
 * @param {BigInt.BigInteger} num 
 */
function numberTo32BytesLE(num) {
    const length = 32;
    const hex = num.toString(16).padStart(length * 2, '0');
    return hexToBytes(hex).reverse(); // reverse for LE
}
/**
 * @param {Uint8Array} uint8a 
 * @returns {BigInt.BigInteger}
 */
function bytesToNumberLE(uint8a) {
    if (!(uint8a instanceof Uint8Array)) throw new Error('Expected Uint8Array');
    return BigInt('0x' + bytesToHex(Uint8Array.from(uint8a).reverse()));
  }

/**
* @param {string} hex 
* @returns {Uint8Array}
*/
// Caching slows it down 2-3x
function hexToBytes(hex) {
    if (typeof hex !== 'string') {
      throw new TypeError('hexToBytes: expected string, got ' + typeof hex);
    }
    if (hex.length % 2) throw new Error('hexToBytes: received invalid unpadded hex');
    var array = Uint8Array.from(Buffer.from(hex, 'hex'));
    if (array.length == 0) throw new Error('hexToBytes: empty array')
    return array;
}

/**
 * @param {Uint8Array} uint8a 
 * @returns {string}
 */
function bytesToHex(uint8a) {
    return Buffer.from(uint8a).toString('hex');
}



/**
 * @param {BigInt.BigInteger} m
 * @param {C25519Key} key
 */
 function signEc(m, key)
 {
     var k, r, s;
     do {
         do {
             k = Utils.random(BigInt.one, C25519Point.n.subtract(BigInt.one));
             r = C25519Point.g.times(k).x.mod(C25519Point.n);
         } while (r.equals(BigInt.zero));
         var kInv = k.modInv(C25519Point.n);
         s = key.x.times(r).add(m).mod(C25519Point.n).times(kInv).mod(C25519Point.n)
     } while (s.equals(BigInt.zero));
     return [r, s];
 }

/**
 * @param {BigInt.BigInteger} m
 * @param {BigInt.BigInteger} r
 * @param {BigInt.BigInteger} s
 * @param {C25519Key} key
 */
function verifyEc(m, r, s, key) {
    console.log("Using the veirfy");
    var w = s.modInv(C25519Point.n);
    var u1 = m.times(w).mod(C25519Point.n);
    var u2 = r.times(w).mod(C25519Point.n);
    var v = C25519Point.g.times(u1).add(key.y.times(u2)).x.mod(C25519Point.n);


    return v.equals(r);
}

/**
 * 
 * @param {BigInt.BigInteger} a 
 * @param {BigInt.BigInteger} b 
 * @returns 
 */
function mod(a, b = C25519Point.p) {
    const res = a.mod(b);
    return res >= BigInt.zero ? res : b.add(res);
}

/**
 * 
 * @param {Uint8Array} encodedPoint 
 * @returns {C25519Point}
 */
// Converts Uint8Array to Point.
// Uses algo from RFC8032 5.1.3.
function fromRawBytes(encodedPoint){
    const normed = encodedPoint.slice();
    normed[31] = encodedPoint[31] & ~0x80;
    const y = bytesToNumberLE(normed);

    return
}

/** @param {Uint8Array} message */
function getM(message)
{
    const size = Math.ceil(C25519Point.n.bitLength().valueOf() / 8);
    return getBig(Hash.shaBuffer(message).slice(0, size));
}

/**
 * @param  {...Uint8Array} arrays 
 * @returns {Uint8Array}
 */
function concatBytes(...arrays) {
    if (!arrays.every((a) => a instanceof Uint8Array)) throw new Error('Expected Uint8Array list');
    if (arrays.length === 1) return arrays[0];
    const length = arrays.reduce((a, arr) => a + arr.length, 0);
    const result = new Uint8Array(length);
    for (let i = 0, pad = 0; i < arrays.length; i++) {
      const arr = arrays[i];
      result.set(arr, pad);
      pad += arr.length;
    }
    return result;
  }