// Tide Protocol - Infrastructure for the Personal Data economy
// Copyright (C) 2019 Tide Foundation Ltd
// 
// This program is free software and is subject to the terms of 
// the Tide Community Open Source License as published by the 
// Tide Foundation Limited. You may modify it and redistribute 
// it in accordance with and subject to the terms of that License.
// This program is distributed WITHOUT WARRANTY of any kind, 
// including without any implied warranty of MERCHANTABILITY or 
// FITNESS FOR A PARTICULAR PURPOSE.
// See the Tide Community Open Source License for more details.
// You should have received a copy of the Tide Community Open 
// Source License along with this program.
// If not, see https://tide.org/licenses_tcosl-1-0-en

import BigInt from 'big-integer';
import Utils from './utils';
import { getBig, getArray } from './bnInput';
import { sign, verify } from './ecSignBase';

/** @template { import('./ecPointBase').default } T
 * @template {EcKeyBase<T,U>} U */
export default class EcKeyBase {
    /**
     * @param {import('./bnInput').BNInput} x
     * @param {T} y
     */
    constructor(x, y) {
        this.x = getBig(x);
        this.y = y;
    }

    get g() { return new this._point(this.curve.g); }

    /**
     * @protected
     * @abstract
     * @return {{ new(point: import('elliptic').curve.base.BasePoint): T ;}}
     */
    get _point() {
        throw new Error('must be implemented by subclass!');
    };

    /**
     * @protected
     * @abstract
     * @return {{ new(x: import('./bnInput').BNInput, y: T): U ;}}
     */
    get _new() {
        throw new Error('must be implemented by subclass!');
    };

    /**
     * @abstract
     * @return { import('./ecCurve').default }
     */
    get curve() {
        throw new Error('must be implemented by subclass!');
    };

    /** @returns {U} */
    public() {
        return new this._new(0, this.y);
    }

    /** @param {string | Uint8Array} message */
    sign(message) {
        return sign(message, this.x, this.curve.nCasted, this.g);
    }

    /**
     * @param {string | Uint8Array} message
     * @param {Uint8Array} signature
     */
    verify(message, signature) {
        return verify(message, signature, this.y, this.curve.nCasted, this.g);
    }

    toString() {
        return Buffer.from(this.toArray()).toString("base64");
    }

    toArray() {
        if (!this.x || !this.y)
            throw new Error('X or y must not be empty');

        var xBuffer = getArray(this.x);
        var yxBuffer = getArray(this.y.x);
        var yyBuffer = getArray(this.y.y);

        if (xBuffer.length > 32 || yxBuffer.length > 32 || yyBuffer.length > 32)
            throw new Error('invalid key format');

        var buffer = new Uint8Array(96);
        buffer.set(xBuffer, 32 - xBuffer.length);
        buffer.set(yxBuffer, 64 - yxBuffer.length);
        buffer.set(yyBuffer, 96 - yyBuffer.length);

        return buffer;
    }

    /**
     * @param {{ new(x: import('./bnInput').BNInput, y: T): U ;}} type 
     * @param {string | Uint8Array} data
     * @template { import('./ecPointBase').default } T
     * @template {EcKeyBase<T,U>} U
     * @protected */
    static _from(type, data) {
        if (!data || data.length <= 0)
            throw new Error('data must not be null or empty');

        const buffer = typeof data === "string" ? Buffer.from(data, "base64") : data;
        if (buffer.length != 96)
            throw new Error('invalid key format');

        const x = getBig(buffer.slice(0, 32));
        const yx = getBig(buffer.slice(32, 64));
        const yy = getBig(buffer.slice(64, 96));
        
        const key = new type(x, null);
        key.y = new key._point(key.curve.point(yx, yy));

        return key;
    }

    /**
     * @param {{ new(x: import('./bnInput').BNInput, y: T): U ;}} type 
     * @param {import('./bnInput').BNInput} x
     * @param {boolean} noPublic
     * @template { import('./ecPointBase').default } T
     * @template {EcKeyBase<T,U>} U
     * @protected */
    static _private(type, x, noPublic = false) {
        const key = new type(x, null);
        key.y = noPublic ? new key._point(key.curve.point())
            : new key._point(key.g.times(x)._value);
        
        return key;
    }

    /**
     * @param {{ new(x: import('./bnInput').BNInput, y: T): U ;}} type 
     * @template { import('./ecPointBase').default } T
     * @template {EcKeyBase<T,U>} U
     * @protected
     * */
    static _generate(type) {
        const key = new type(0, null);

        key.x = Utils.random(BigInt.one, key.curve.nCasted.subtract(BigInt.one));
        key.y = new key._point(key.g.times(key.x)._value);

        return key;
    }
}
