// Tide Protocol - Infrastructure for the Personal Data economy
// Copyright (C) 2019 Tide Foundation Ltd
// 
// This program is free software and is subject to the terms of 
// the Tide Community Open Source License as published by the 
// Tide Foundation Limited. You may modify it and redistribute 
// it in accordance with and subject to the terms of that License.
// This program is distributed WITHOUT WARRANTY of any kind, 
// including without any implied warranty of MERCHANTABILITY or 
// FITNESS FOR A PARTICULAR PURPOSE.
// See the Tide Community Open Source License for more details.
// You should have received a copy of the Tide Community Open 
// Source License along with this program.
// If not, see https://tide.org/licenses_tcosl-1-0-en

import BigInt from 'big-integer';
import Hash from './hash';
import { getBig, getArray } from './bnInput';
import Utils from './utils';
import EcPointBase from './ecPointBase';

/**
 * @param {Uint8Array|string} message 
 * @param {BigInt.BigInteger} x
 * @param {BigInt.BigInteger} n
 * @param {EcPointBase} g
 * @returns {Uint8Array}
 */
export function sign(message, x, n, g) {
    var buff = typeof message === 'string' ? Buffer.from(message, 'utf-8') : message;
    var [r, s] = signBase(getM(buff, n), x, n, g);
    return map(r, s);
}

/**
 * @param {Uint8Array|string} message
 * @param {Uint8Array} signature
 * @param {EcPointBase} y
 * @param {BigInt.BigInteger} n
 * @param {EcPointBase} g
 */
export function verify(message, signature, y, n, g) {
    if (!signature || signature.length != 64)
        return false;

    var buff = typeof message === 'string' ? Buffer.from(message, 'utf-8') : message;
    var [r, s] = unmap(signature);
    return verifyBase(getM(buff, n), r, s, n, y, g);
}

/**
 * @param {BigInt.BigInteger} r
 * @param {BigInt.BigInteger} s
 */
export function map(r, s) {
    var bufferR = getArray(r);
    var bufferS = getArray(s);

    var buffer = Buffer.alloc(64);
    buffer.set(bufferR, 32 - bufferR.length);
    buffer.set(bufferS, 64 - bufferS.length);

    return buffer;
}
/** @param {Uint8Array} signature */
export function unmap(signature) {
    return [getBig(signature.slice(0, 32)), getBig(signature.slice(32))];
}

/**
 * @param {BigInt.BigInteger} m
 * @param {BigInt.BigInteger} x
 * @param {BigInt.BigInteger} n
 * @param {EcPointBase} g
 */
function signBase(m, x, n, g)
{
    var k, r, s;
    do {
        do {
            k = Utils.random(BigInt.one, n.subtract(BigInt.one));
            r = g.times(k).x.mod(n);
        } while (r.equals(BigInt.zero));
        var kInv = k.modInv(n);
        s = x.times(r).add(m).mod(n).times(kInv).mod(n)
    } while (s.equals(BigInt.zero));
    return [r, s];
}

/**
 * @param {BigInt.BigInteger} m
 * @param {BigInt.BigInteger} r
 * @param {BigInt.BigInteger} s
 * @param {BigInt.BigInteger} n
 * @param {EcPointBase} y
 * @param {EcPointBase} g
 */
function verifyBase(m, r, s, n, y, g) {
    var w = s.modInv(n);
    var u1 = m.times(w).mod(n);
    var u2 = r.times(w).mod(n);
    var v = g.times(u1).add(y.times(u2)).x.mod(n);

    return v.equals(r);
}

/** @param {Uint8Array} message
 * @param {BigInt.BigInteger} n */
function getM(message, n)
{
    const size = Math.ceil(n.bitLength().valueOf() / 8);
    return getBig(Hash.shaBuffer(message).slice(0, size));
}
